@@ -54,8 +54,10 @@
 #include "device.h"
 #include "manager.h"
 #include "error.h"
+#include "telephony.h"
 #include "headset.h"
 #include "glib-helper.h"
+#include "dbus-common.h"
 
 #define DC_TIMEOUT 3000
 
@@ -66,20 +68,31 @@
 #define HEADSET_GAIN_SPEAKER 'S'
 #define HEADSET_GAIN_MICROPHONE 'M'
 
-#define AG_FEATURE_THREE_WAY_CALLING             0x0001
-#define AG_FEATURE_EC_ANDOR_NR                   0x0002
-#define AG_FEATURE_VOICE_RECOGNITION             0x0004
-#define AG_FEATURE_INBAND_RINGTONE               0x0008
-#define AG_FEATURE_ATTACH_NUMBER_TO_VOICETAG     0x0010
-#define AG_FEATURE_REJECT_A_CALL                 0x0020
-#define AG_FEATURE_ENHANCES_CALL_STATUS          0x0040
-#define AG_FEATURE_ENHANCES_CALL_CONTROL         0x0080
-#define AG_FEATURE_EXTENDED_ERROR_RESULT_CODES   0x0100
-
-static uint32_t ag_features = 0;
+static struct {
+	gboolean telephony_ready;	/* Telephony plugin initialized */
+	uint32_t features;		/* HFP AG features */
+	const struct indicator *indicators;	/* Available HFP indicators */
+	int er_mode;			/* Event reporting mode */
+	int er_ind;			/* Event reporting for indicators */
+	int rh;				/* Response and Hold state */
+	char *number;			/* Incoming phone number */
+	int number_type;		/* Incoming number type */
+	guint ring_timer;		/* For incoming call indication */
+} ag = {
+	.telephony_ready = FALSE,
+	.features = 0,
+	.er_mode = 3,
+	.er_ind = 0,
+	.rh = -1,
+	.number = NULL,
+	.number_type = 0,
+	.ring_timer = 0,
+};
 
 static gboolean sco_hci = TRUE;
 
+static GSList *active_devices = NULL;
+
 static char *str_state[] = {
 	"HEADSET_STATE_DISCONNECTED",
 	"HEADSET_STATE_CONNECT_IN_PROGRESS",
@@ -116,8 +129,6 @@ struct headset {
 
 	gboolean auto_dc;
 
-	guint ring_timer;
-
 	guint dc_timer;
 
 	char buf[BUF_SIZE];
@@ -127,8 +138,8 @@ struct headset {
 	gboolean hfp_active;
 	gboolean search_hfp;
 	gboolean cli_active;
-	char *ph_number;
-	int type;
+	gboolean cme_enabled;
+	gboolean cwa_enabled;
 
 	headset_state_t state;
 	struct pending_connect *pending;
@@ -136,7 +147,7 @@ struct headset {
 	int sp_gain;
 	int mic_gain;
 
-	unsigned int hfp_features;
+	unsigned int hf_features;
 	headset_lock_t lock;
 };
 
@@ -145,6 +156,12 @@ struct event {
 	int (*callback) (struct audio_device *device, const char *buf);
 };
 
+static inline DBusMessage *invalid_args(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+			"Invalid arguments in method call");
+}
+
 static DBusHandlerResult error_not_supported(DBusConnection *conn,
 							DBusMessage *msg)
 {
@@ -165,16 +182,85 @@ static int rfcomm_connect(struct audio_device *device, headset_stream_cb_t cb,
 static int get_records(struct audio_device *device, headset_stream_cb_t cb,
 			void *user_data, unsigned int *cb_id);
 
-static int headset_send(struct headset *hs, char *format, ...)
+static void print_ag_features(uint32_t features)
+{
+	GString *gstr;
+	char *str;
+
+	if (features == 0) {
+		debug("HFP AG features: (none)");
+		return;
+	}
+
+	gstr = g_string_new("HFP AG features: ");
+
+	if (features & AG_FEATURE_THREE_WAY_CALLING)
+		g_string_append(gstr, "\"Three-way calling\" ");
+	if (features & AG_FEATURE_EC_ANDOR_NR)
+		g_string_append(gstr, "\"EC and/or NR function\" ");
+	if (features & AG_FEATURE_VOICE_RECOGNITION)
+		g_string_append(gstr, "\"Voice recognition function\" ");
+	if (features & AG_FEATURE_INBAND_RINGTONE)
+		g_string_append(gstr, "\"In-band ring tone capability\" ");
+	if (features & AG_FEATURE_ATTACH_NUMBER_TO_VOICETAG)
+		g_string_append(gstr, "\"Attach a number to a voice tag\" ");
+	if (features & AG_FEATURE_REJECT_A_CALL)
+		g_string_append(gstr, "\"Ability to reject a call\" ");
+	if (features & AG_FEATURE_ENHANCED_CALL_STATUS)
+		g_string_append(gstr, "\"Enhanced call status\" ");
+	if (features & AG_FEATURE_ENHANCED_CALL_CONTROL)
+		g_string_append(gstr, "\"Enhanced call control\" ");
+	if (features & AG_FEATURE_EXTENDED_ERROR_RESULT_CODES)
+		g_string_append(gstr, "\"Extended Error Result Codes\" ");
+
+	str = g_string_free(gstr, FALSE);
+
+	debug("%s", str);
+
+	g_free(str);
+}
+
+static void print_hf_features(uint32_t features)
+{
+	GString *gstr;
+	char *str;
+
+	if (features == 0) {
+		debug("HFP HF features: (none)");
+		return;
+	}
+
+	gstr = g_string_new("HFP HF features: ");
+
+	if (features & HF_FEATURE_EC_ANDOR_NR)
+		g_string_append(gstr, "\"EC and/or NR function\" ");
+	if (features & HF_FEATURE_CALL_WAITING_AND_3WAY)
+		g_string_append(gstr, "\"Call waiting and 3-way calling\" ");
+	if (features & HF_FEATURE_CLI_PRESENTATION)
+		g_string_append(gstr, "\"CLI presentation capability\" ");
+	if (features & HF_FEATURE_VOICE_RECOGNITION)
+		g_string_append(gstr, "\"Voice recognition activation\" ");
+	if (features & HF_FEATURE_REMOTE_VOLUME_CONTROL)
+		g_string_append(gstr, "\"Remote volume control\" ");
+	if (features & HF_FEATURE_ENHANCED_CALL_STATUS)
+		g_string_append(gstr, "\"Enhanced call status\" ");
+	if (features & HF_FEATURE_ENHANCED_CALL_CONTROL)
+		g_string_append(gstr, "\"Enhanced call control\" ");
+
+	str = g_string_free(gstr, FALSE);
+
+	debug("%s", str);
+
+	g_free(str);
+}
+
+static int headset_send_valist(struct headset *hs, char *format, va_list ap)
 {
 	char rsp[BUF_SIZE];
-	va_list ap;
 	ssize_t total_written, written, count;
 	int fd;
 
-	va_start(ap, format);
 	count = vsnprintf(rsp, sizeof(rsp), format, ap);
-	va_end(ap);
 
 	if (count < 0)
 		return -EINVAL;
@@ -198,6 +284,18 @@ static int headset_send(struct headset *hs, char *format, ...)
 	return 0;
 }
 
+static int headset_send(struct headset *hs, char *format, ...)
+{
+	va_list ap;
+	int ret;
+
+	va_start(ap, format);
+	ret = headset_send_valist(hs, format, ap);
+	va_end(ap);
+
+	return ret;
+}
+
 static int supported_features(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
@@ -206,24 +304,76 @@ static int supported_features(struct audio_device *device, const char *buf)
 	if (strlen(buf) < 9)
 		return -EINVAL;
 
-	hs->hfp_features = strtoul(&buf[8], NULL, 10);
-	err = headset_send(hs, "\r\n+BRSF=%u\r\n", ag_features);
+	hs->hf_features = strtoul(&buf[8], NULL, 10);
+
+	print_hf_features(hs->hf_features);
+
+	err = headset_send(hs, "\r\n+BRSF=%u\r\n", ag.features);
 	if (err < 0)
 		return err;
 
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
+static char *indicator_ranges(const struct indicator *indicators)
+{
+	int i;
+	GString *gstr;
+
+	gstr = g_string_new("\r\n+CIND:");
+
+	for (i = 0; indicators[i].desc != NULL; i++) {
+		if (i == 0)
+			g_string_append_printf(gstr, "(\"%s\",(%s))",
+						indicators[i].desc,
+						indicators[i].range);
+		else
+			g_string_append_printf(gstr, ",(\"%s\",(%s))",
+						indicators[i].desc,
+						indicators[i].range);
+	}
+
+	g_string_append(gstr, "\r\n");
+
+	return g_string_free(gstr, FALSE);
+}
+
+static char *indicator_values(const struct indicator *indicators)
+{
+	int i;
+	GString *gstr;
+
+	gstr = g_string_new("\r\n+CIND:");
+
+	for (i = 0; indicators[i].desc != NULL; i++) {
+		if (i == 0)
+			g_string_append_printf(gstr, "%d", indicators[i].val);
+		else
+			g_string_append_printf(gstr, ",%d", indicators[i].val);
+	}
+
+	g_string_append(gstr, "\r\n");
+
+	return g_string_free(gstr, FALSE);
+}
+
 static int report_indicators(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
 	int err;
+	char *str;
+
+	if (strlen(buf) < 8)
+		return -EINVAL;
 
 	if (buf[7] == '=')
-		err = headset_send(hs, "\r\n+CIND:(\"service\",(0,1)),"
-				"(\"call\",(0,1)),(\"callsetup\",(0-3))\r\n");
+		str = indicator_ranges(ag.indicators);
 	else
-		err = headset_send(hs, "\r\n+CIND:1,0,0\r\n");
+		str = indicator_values(ag.indicators);
+
+	err = headset_send(hs, str);
+
+	g_free(str);
 
 	if (err < 0)
 		return err;
@@ -330,7 +480,7 @@ static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	sk = g_io_channel_unix_get_fd(chan);
 
-	info("SCO fd=%d", sk);
+	debug("SCO fd=%d", sk);
 	hs->sco = chan;
 	p->io = NULL;
 
@@ -359,7 +509,7 @@ static int sco_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	err = bt_sco_connect(&dev->src, &dev->dst, sco_connect_cb, dev);
 	if (err < 0) {
 		error("connect: %s (%d)", strerror(-err), -err);
-		return -err;
+		return err;
 	}
 
 	headset_set_state(dev, HEADSET_STATE_PLAY_IN_PROGRESS);
@@ -404,11 +554,29 @@ static void hfp_slc_complete(struct audio_device *dev)
 		pending_connect_finalize(dev);
 }
 
-static int event_reporting(struct audio_device *dev, const char *buf)
+static int telephony_generic_rsp(struct audio_device *device, cme_error_t err)
 {
-	struct headset *hs = dev->headset;
+	struct headset *hs = device->headset;
+
+	if (err != CME_ERROR_NONE) {
+		if (hs->cme_enabled)
+			return headset_send(hs, "\r\n+CME ERROR:%d\r\n", err);
+		else
+			return headset_send(hs, "\r\nERROR\r\n");
+	}
+
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+int telephony_event_reporting_rsp(void *telephony_device, cme_error_t err)
+{
+	struct audio_device *device = telephony_device;
+	struct headset *hs = device->headset;
 	int ret;
 
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
+
 	ret = headset_send(hs, "\r\nOK\r\n");
 	if (ret < 0)
 		return ret;
@@ -416,10 +584,44 @@ static int event_reporting(struct audio_device *dev, const char *buf)
 	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
 		return 0;
 
-	if (ag_features & AG_FEATURE_THREE_WAY_CALLING)
+	if (ag.features & AG_FEATURE_THREE_WAY_CALLING)
 		return 0;
 
-	hfp_slc_complete(dev);
+	hfp_slc_complete(device);
+
+	return 0;
+}
+
+static int event_reporting(struct audio_device *dev, const char *buf)
+{
+	char **tokens; /* <mode>, <keyp>, <disp>, <ind>, <bfr> */
+
+	if (strlen(buf) < 13)
+		return -EINVAL;
+
+	tokens = g_strsplit(&buf[8], ",", 5);
+	if (g_strv_length(tokens) < 4) {
+		g_strfreev(tokens);
+		return -EINVAL;
+	}
+
+	ag.er_mode = atoi(tokens[0]);
+	ag.er_ind = atoi(tokens[3]);
+
+	g_strfreev(tokens);
+	tokens = NULL;
+
+	debug("Event reporting (CMER): mode=%d, ind=%d",
+			ag.er_mode, ag.er_ind);
+
+	switch (ag.er_ind) {
+	case 0:
+	case 1:
+		telephony_event_reporting_req(dev, ag.er_ind);
+		break;
+	default:
+		return -EINVAL;
+	}
 
 	return 0;
 }
@@ -445,68 +647,86 @@ static int call_hold(struct audio_device *dev, const char *buf)
 	return 0;
 }
 
-static int answer_call(struct audio_device *device, const char *buf)
+static int button_press(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
-	int err;
 
 	g_dbus_emit_signal(device->conn, device->path,
 			AUDIO_HEADSET_INTERFACE, "AnswerRequested",
 			DBUS_TYPE_INVALID);
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	}
 
-	if (!hs->hfp_active)
-		return headset_send(hs, "\r\nOK\r\n");
-
-	if (hs->ph_number) {
-		g_free(hs->ph_number);
-		hs->ph_number = NULL;
-	}
-
-	err = headset_send(hs, "\r\nOK\r\n");
-	if (err < 0)
-		return err;
-
-	/*+CIEV: (call = 1)*/
-	err = headset_send(hs, "\r\n+CIEV:2,1\r\n");
-	if (err < 0)
-		return err;
-
-	/*+CIEV: (callsetup = 0)*/
-	return headset_send(hs, "\r\n+CIEV:3,0\r\n");
+	return headset_send(hs, "\r\nOK\r\n");
 }
 
-static int terminate_call(struct audio_device *device, const char *buf)
+int telephony_answer_call_rsp(void *telephony_device, cme_error_t err)
 {
+	struct audio_device *device = telephony_device;
 	struct headset *hs = device->headset;
-	int err;
+
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
+
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
+	}
+
+	if (ag.number) {
+		g_free(ag.number);
+		ag.number = NULL;
+	}
+
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+int telephony_anser_call_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int answer_call(struct audio_device *device, const char *buf)
+{
+	telephony_answer_call_req(device);
+
+	return 0;
+}
+
+int telephony_terminate_call_rsp(void *telephony_device,
+					cme_error_t err)
+{
+	struct audio_device *device = telephony_device;
+	struct headset *hs = device->headset;
+
+	if (err != CME_ERROR_NONE)
+		return telephony_generic_rsp(telephony_device, err);
 
 	g_dbus_emit_signal(device->conn, device->path,
 			AUDIO_HEADSET_INTERFACE, "CallTerminated",
 			DBUS_TYPE_INVALID);
 
-	err = headset_send(hs, "\r\nOK\r\n");
-	if (err < 0)
-		return err;
-
-	if (hs->ph_number) {
-		g_free(hs->ph_number);
-		hs->ph_number = NULL;
+	if (ag.number) {
+		g_free(ag.number);
+		ag.number = NULL;
 	}
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
-		/*+CIEV: (callsetup = 0)*/
-		return headset_send(hs, "\r\n+CIEV:3,0\r\n");
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
 	}
 
-	/*+CIEV: (call = 0)*/
-	return headset_send(hs, "\r\n+CIEV:2,0\r\n");
+	return headset_send(hs, "\r\nOK\r\n");
+}
+
+static int terminate_call(struct audio_device *device, const char *buf)
+{
+	telephony_terminate_call_req(device);
+
+	return 0;
 }
 
 static int cli_notification(struct audio_device *device, const char *buf)
@@ -521,9 +741,56 @@ static int cli_notification(struct audio_device *device, const char *buf)
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
+int telephony_response_and_hold_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int response_and_hold(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 8)
+		return -EINVAL;
+
+	if (buf[7] == '=')
+		telephony_response_and_hold_req(device, atoi(&buf[8]) < 0);
+	else if (ag.rh >= 0)
+		return headset_send(hs, "\r\n+BTRH:%d\r\n", ag.rh);
+	else
+		return headset_send(hs, "\r\nOK\r\n", ag.rh);
+
+	return 0;
+}
+
+int telephony_last_dialed_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int last_dialed_number(struct audio_device *device, const char *buf)
+{
+	telephony_last_dialed_number_req(device);
+
+	return 0;
+}
+
+int telephony_dial_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int dial_number(struct audio_device *device, const char *buf)
+{
+	telephony_dial_number_req(device, &buf[3]);
+
+	return 0;
+}
+
 static int signal_gain_setting(struct audio_device *device, const char *buf)
 {
 	struct headset *hs = device->headset;
+	const char *property;
 	const char *name;
 	dbus_uint16_t gain;
 
@@ -544,12 +811,14 @@ static int signal_gain_setting(struct audio_device *device, const char *buf)
 		if (hs->sp_gain == gain)
 			goto ok;
 		name = "SpeakerGainChanged";
+		property = "SpeakerGain";
 		hs->sp_gain = gain;
 		break;
 	case HEADSET_GAIN_MICROPHONE:
 		if (hs->mic_gain == gain)
 			goto ok;
 		name = "MicrophoneGainChanged";
+		property = "MicrophoneGain";
 		hs->mic_gain = gain;
 		break;
 	default:
@@ -558,24 +827,113 @@ static int signal_gain_setting(struct audio_device *device, const char *buf)
 	}
 
 	g_dbus_emit_signal(device->conn, device->path,
-				    AUDIO_HEADSET_INTERFACE, name,
-				    DBUS_TYPE_UINT16, &gain,
-				    DBUS_TYPE_INVALID);
+				AUDIO_HEADSET_INTERFACE, name,
+				DBUS_TYPE_UINT16, &gain,
+				DBUS_TYPE_INVALID);
+
+	dbus_connection_emit_property_changed(device->conn, device->path,
+				AUDIO_HEADSET_INTERFACE, property,
+				DBUS_TYPE_UINT16, &gain);
 
 ok:
 	return headset_send(hs, "\r\nOK\r\n");
 }
 
+int telephony_transmit_dtmf_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int dtmf_tone(struct audio_device *device, const char *buf)
+{
+	if (strlen(buf) < 8) {
+		error("Too short string for DTMF tone");
+		return -EINVAL;
+	}
+
+	telephony_transmit_dtmf_req(device, buf[7]);
+
+	return 0;
+}
+
+int telephony_subscriber_number_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int subscriber_number(struct audio_device *device, const char *buf)
+{
+	telephony_subscriber_number_req(device);
+
+	return 0;
+}
+
+int telephony_list_current_calls_rsp(void *telephony_device, cme_error_t err)
+{
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int list_current_calls(struct audio_device *device, const char *buf)
+{
+	telephony_list_current_calls_req(device);
+
+	return 0;
+}
+
+static int extended_errors(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] == '1') {
+		hs->cme_enabled = TRUE;
+		debug("CME errors enabled for headset %p", hs);
+	} else {
+		hs->cme_enabled = FALSE;
+		debug("CME errors disabled for headset %p", hs);
+	}
+
+	return 0;
+}
+
+static int call_waiting_notify(struct audio_device *device, const char *buf)
+{
+	struct headset *hs = device->headset;
+
+	if (strlen(buf) < 9)
+		return -EINVAL;
+
+	if (buf[8] == '1') {
+		hs->cwa_enabled = TRUE;
+		debug("Call waiting notifiaction enabled for headset %p", hs);
+	} else {
+		hs->cwa_enabled = FALSE;
+		debug("Call waiting notification disabled for headset %p", hs);
+	}
+
+	return 0;
+}
+
 static struct event event_callbacks[] = {
 	{ "ATA", answer_call },
+	{ "ATD", dial_number },
 	{ "AT+VG", signal_gain_setting },
 	{ "AT+BRSF", supported_features },
 	{ "AT+CIND", report_indicators },
 	{ "AT+CMER", event_reporting },
 	{ "AT+CHLD", call_hold },
 	{ "AT+CHUP", terminate_call },
-	{ "AT+CKPD", answer_call },
+	{ "AT+CKPD", button_press },
 	{ "AT+CLIP", cli_notification },
+	{ "AT+BTRH", response_and_hold },
+	{ "AT+BLDN", last_dialed_number },
+	{ "AT+VTS", dtmf_tone },
+	{ "AT+CNUM", subscriber_number },
+	{ "AT+CLCC", list_current_calls },
+	{ "AT+CMEE", extended_errors },
+	{ "AT+CCWA", call_waiting_notify },
 	{ 0 }
 };
 
@@ -826,7 +1184,6 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 	}
 
 	sdp_list_free(classes, free);
-	sdp_record_free(record);
 
 	return;
 
@@ -839,8 +1196,6 @@ failed_not_supported:
 failed:
 	if (classes)
 		sdp_list_free(classes, free);
-	if (record)
-		sdp_record_free(record);
 	pending_connect_finalize(dev);
 	headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 }
@@ -1006,6 +1361,10 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 						".AlreadyConnected",
 						"Already Connected");
 
+	if (hs->hfp_handle && !ag.telephony_ready)
+		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotReady",
+					"Telephony subsystem not ready");
+
 	err = rfcomm_connect(device, NULL, NULL, NULL);
 	if (err < 0)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
@@ -1019,26 +1378,57 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 	return NULL;
 }
 
+static int hfp_cmp(struct headset *hs)
+{
+	if (hs->hfp_active)
+		return 0;
+	else
+		return -1;
+}
+
+static void send_foreach_headset(GSList *devices,
+					int (*cmp)(struct headset *hs),
+					char *format, ...)
+{
+	GSList *l;
+	va_list ap;
+
+	for (l = devices; l != NULL; l = l->next) {
+		struct audio_device *device = l->data;
+		struct headset *hs = device->headset;
+		int ret;
+
+		if (cmp && cmp(hs) != 0)
+			continue;
+
+		va_start(ap, format);
+		ret = headset_send_valist(hs, format, ap);
+		if (ret < 0)
+			error("Failed to send to headset: %s (%d)",
+					strerror(-ret), -ret);
+		va_end(ap);
+	}
+}
+
+static int cli_cmp(struct headset *hs)
+{
+	if (!hs->hfp_active)
+		return -1;
+
+	if (hs->cli_active)
+		return 0;
+	else
+		return -1;
+}
+
 static gboolean ring_timer_cb(gpointer data)
 {
-	struct audio_device *device = data;
-	struct headset *hs = device->headset;
-	int err;
+	send_foreach_headset(active_devices, NULL, "\r\nRING\r\n");
 
-	err = headset_send(hs, "\r\nRING\r\n");
-
-	if (err < 0)
-		error("Error while sending RING: %s (%d)",
-				strerror(-err), -err);
-
-	if (hs->cli_active && hs->ph_number) {
-		err = headset_send(hs, "\r\n+CLIP:\"%s\",%d\r\n",
-					hs->ph_number, hs->type);
-
-		if (err < 0)
-			error("Error while sending CLIP: %s (%d)",
-					strerror(-err), -err);
-	}
+	if (ag.number)
+		send_foreach_headset(active_devices, cli_cmp,
+					"\r\n+CLIP:\"%s\",%d\r\n",
+					ag.number, ag.number_type);
 
 	return TRUE;
 }
@@ -1060,7 +1450,7 @@ static DBusMessage *hs_ring(DBusConnection *conn, DBusMessage *msg,
 	if (!reply)
 		return NULL;
 
-	if (hs->ring_timer) {
+	if (ag.ring_timer) {
 		debug("IndicateCall received when already indicating");
 		goto done;
 	}
@@ -1072,24 +1462,14 @@ static DBusMessage *hs_ring(DBusConnection *conn, DBusMessage *msg,
 						"%s", strerror(-err));
 	}
 
-	if (hs->cli_active && hs->ph_number) {
-		err = headset_send(hs, "\r\n+CLIP:\"%s\",%d\r\n",
-					hs->ph_number, hs->type);
-		if (err < 0) {
-			dbus_message_unref(reply);
-			return g_dbus_create_error(msg, ERROR_INTERFACE
-						".Failed", "%s",
-						strerror(-err));
-		}
-	}
-
-	hs->ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, device);
+	ring_timer_cb(NULL);
+	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
 
 done:
 	return reply;
 }
 
-static DBusMessage *hs_cancel_ringing(DBusConnection *conn,
+static DBusMessage *hs_cancel_call(DBusConnection *conn,
 					DBusMessage *msg,
 					void *data)
 {
@@ -1106,26 +1486,11 @@ static DBusMessage *hs_cancel_ringing(DBusConnection *conn,
 	if (!reply)
 		return NULL;
 
-	if (!hs->ring_timer) {
-		debug("Got CancelRinging method call but ringing is not in progress");
-		goto done;
-	}
-
-	g_source_remove(hs->ring_timer);
-	hs->ring_timer = 0;
-
-done:
-	if (hs->hfp_active) {
-		int err;
-		/*+CIEV: (callsetup = 0)*/
-		err = headset_send(hs, "\r\n+CIEV:3,0\r\n");
-		if (err < 0) {
-			dbus_message_unref(reply);
-			return g_dbus_create_error(msg, ERROR_INTERFACE
-						".Failed", "%s",
-						strerror(-err));
-		}
-	}
+	if (ag.ring_timer) {
+		g_source_remove(ag.ring_timer);
+		ag.ring_timer = 0;
+	} else
+		debug("Got CancelCall method call but no call is active");
 
 	return reply;
 }
@@ -1225,12 +1590,12 @@ static DBusMessage *hs_get_mic_gain(DBusConnection *conn,
 
 static DBusMessage *hs_set_gain(DBusConnection *conn,
 				DBusMessage *msg,
-				void *data, char type)
+				void *data, uint16_t gain,
+				char type)
 {
 	struct audio_device *device = data;
 	struct headset *hs = device->headset;
 	DBusMessage *reply;
-	dbus_uint16_t gain;
 	int err;
 
 	if (hs->state < HEADSET_STATE_CONNECTED)
@@ -1238,10 +1603,6 @@ static DBusMessage *hs_set_gain(DBusConnection *conn,
 						".NotConnected",
 						"Device not Connected");
 
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
-				DBUS_TYPE_INVALID))
-		return NULL;
-
 	if (gain > 15)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".InvalidArgument",
@@ -1285,127 +1646,152 @@ static DBusMessage *hs_set_speaker_gain(DBusConnection *conn,
 					DBusMessage *msg,
 					void *data)
 {
-	return hs_set_gain(conn, msg, data, HEADSET_GAIN_SPEAKER);
+	uint16_t gain;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
+				DBUS_TYPE_INVALID))
+		return NULL;
+
+	return hs_set_gain(conn, msg, data, gain, HEADSET_GAIN_SPEAKER);
 }
 
 static DBusMessage *hs_set_mic_gain(DBusConnection *conn,
 					DBusMessage *msg,
 					void *data)
 {
-	return hs_set_gain(conn, msg, data, HEADSET_GAIN_MICROPHONE);
-}
+	uint16_t gain;
 
-static DBusMessage *hf_setup_call(DBusConnection *conn,
-					DBusMessage *msg,
-					void *data)
-{
-	struct audio_device *device = data;
-	struct headset *hs = device->headset;
-	DBusMessage *reply;
-	const char *value;
-	int err;
-
-	if (!hs->hfp_active)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotSuppported",
-						"Not Supported");
-
-	if (hs->state < HEADSET_STATE_CONNECTED)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotConnected",
-						"Device not Connected");
-
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &value,
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_UINT16, &gain,
 				DBUS_TYPE_INVALID))
 		return NULL;
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
-	if (!strncmp(value, "incoming", 8))
-		err = headset_send(hs, "\r\n+CIEV:3,1\r\n");
-	else if (!strncmp(value, "outgoing", 8))
-		err = headset_send(hs, "\r\n+CIEV:3,2\r\n");
-	else if (!strncmp(value, "remote", 6))
-		err = headset_send(hs, "\r\n+CIEV:3,3\r\n");
-	else
-		err = -EINVAL;
-
-	if (err < 0) {
-		dbus_message_unref(reply);
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-						"%s", strerror(-err));
-	}
-
-	return reply;
+	return hs_set_gain(conn, msg, data, gain, HEADSET_GAIN_MICROPHONE);
 }
 
-static DBusMessage *hf_identify_call(DBusConnection *conn,
-						DBusMessage *msg,
-						void *data)
+static DBusMessage *hs_get_properties(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
 	struct audio_device *device = data;
-	struct headset *hs = device->headset;
 	DBusMessage *reply;
-	const char *number;
-	dbus_int32_t type;
-
-	if (!hs->hfp_active && !hs->cli_active)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotSuppported",
-						"Not Supported");
-
-	if (hs->state < HEADSET_STATE_CONNECTED)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".NotConnected",
-						"Device not Connected");
-
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &number,
-			      DBUS_TYPE_INT32, &type, DBUS_TYPE_INVALID))
-		return NULL;
+	DBusMessageIter iter;
+	DBusMessageIter dict;
+	gboolean value;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
 		return NULL;
 
-	g_free(hs->ph_number);
+	dbus_message_iter_init_append(reply, &iter);
 
-	hs->ph_number = g_strdup(number);
-	hs->type = type;
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+
+	/* Playing */
+	value = (device->headset->state == HEADSET_STATE_PLAYING);
+	dbus_message_iter_append_dict_entry(&dict, "Playing",
+						DBUS_TYPE_BOOLEAN, &value);
+
+	/* Connected */
+	value = (device->headset->state >= HEADSET_STATE_CONNECTED);
+	dbus_message_iter_append_dict_entry(&dict, "Connected",
+						DBUS_TYPE_BOOLEAN, &value);
+
+	if (!value)
+		goto done;
+
+	/* SpeakerGain */
+	dbus_message_iter_append_dict_entry(&dict, "SpeakerGain",
+						DBUS_TYPE_UINT16,
+						&device->headset->sp_gain);
+
+	/* MicrophoneGain */
+	dbus_message_iter_append_dict_entry(&dict, "MicrophoneGain",
+						DBUS_TYPE_UINT16,
+						&device->headset->mic_gain);
+
+done:
+	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
 }
 
+static DBusMessage *hs_set_property(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *property;
+	DBusMessageIter iter;
+	DBusMessageIter sub;
+	uint16_t gain;
+
+	if (!dbus_message_iter_init(msg, &iter))
+		return invalid_args(msg);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
+		return invalid_args(msg);
+
+	dbus_message_iter_get_basic(&iter, &property);
+	dbus_message_iter_next(&iter);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+		return invalid_args(msg);
+	dbus_message_iter_recurse(&iter, &sub);
+
+	if (g_str_equal("SpeakerGain", property)) {
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT16)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &gain);
+		return hs_set_gain(conn, msg, data, gain,
+					HEADSET_GAIN_SPEAKER);
+	} else if (g_str_equal("MicrophoneGain", property)) {
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT16)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &gain);
+		return hs_set_gain(conn, msg, data, gain,
+					HEADSET_GAIN_MICROPHONE);
+	}
+
+	return invalid_args(msg);
+}
 static GDBusMethodTable headset_methods[] = {
 	{ "Connect",		"",	"",	hs_connect,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Disconnect",		"",	"",	hs_disconnect },
 	{ "IsConnected",	"",	"b",	hs_is_connected },
 	{ "IndicateCall",	"",	"",	hs_ring },
-	{ "CancelCall",		"",	"",	hs_cancel_ringing },
+	{ "CancelCall",		"",	"",	hs_cancel_call },
 	{ "Play",		"",	"",	hs_play,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Stop",		"",	"",	hs_stop },
-	{ "IsPlaying",		"",	"b",	hs_is_playing },
-	{ "GetSpeakerGain",	"",	"q",	hs_get_speaker_gain },
-	{ "GetMicrophoneGain",	"",	"q",	hs_get_mic_gain },
-	{ "SetSpeakerGain",	"q",	"",	hs_set_speaker_gain },
-	{ "SetMicrophoneGain",	"q",	"",	hs_set_mic_gain },
-	{ "SetupCall",		"s",	"",	hf_setup_call },
-	{ "IdentifyCall",	"si",	"",	hf_identify_call },
+	{ "IsPlaying",		"",	"b",	hs_is_playing,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetSpeakerGain",	"",	"q",	hs_get_speaker_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetMicrophoneGain",	"",	"q",	hs_get_mic_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "SetSpeakerGain",	"q",	"",	hs_set_speaker_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "SetMicrophoneGain",	"q",	"",	hs_set_mic_gain,
+						G_DBUS_METHOD_FLAG_DEPRECATED },
+	{ "GetProperties",	"",	"a{sv}",hs_get_properties },
+	{ "SetProperty",	"sv",	"",	hs_set_property },
 	{ NULL, NULL, NULL, NULL }
 };
 
 static GDBusSignalTable headset_signals[] = {
-	{ "Connected",			""	},
-	{ "Disconnected",		""	},
+	{ "Connected",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "Disconnected",		"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "AnswerRequested",		""	},
-	{ "Stopped",			""	},
-	{ "Playing",			""	},
-	{ "SpeakerGainChanged",		"q"	},
-	{ "MicrophoneGainChanged",	"q"	},
+	{ "Stopped",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "Playing",			"",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "SpeakerGainChanged",		"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
+	{ "MicrophoneGainChanged",	"q",	G_DBUS_SIGNAL_FLAG_DEPRECATED },
 	{ "CallTerminated",		""	},
+	{ "PropertyChanged",		"sv"	},
 	{ NULL, NULL }
 };
 
@@ -1523,7 +1909,6 @@ struct headset *headset_init(struct audio_device *dev, sdp_record_t *record,
 	hs->search_hfp = server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID);
 	hs->hfp_active = FALSE;
 	hs->cli_active = FALSE;
-	hs->ph_number = NULL;
 
 	if (!record)
 		goto register_iface;
@@ -1562,12 +1947,11 @@ register_iface:
 uint32_t headset_config_init(GKeyFile *config)
 {
 	GError *err = NULL;
-	gboolean value;
 	char *str;
 
 	/* Use the default values if there is no config file */
 	if (config == NULL)
-		return ag_features;
+		return ag.features;
 
 	str = g_key_file_get_string(config, "General", "SCORouting",
 					&err);
@@ -1585,88 +1969,7 @@ uint32_t headset_config_init(GKeyFile *config)
 		g_free(str);
 	}
 
-	value = g_key_file_get_boolean(config, "Headset", "3WayCalling",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_THREE_WAY_CALLING;
-
-	value = g_key_file_get_boolean(config, "Headset", "EchoCancelNoiseCancel",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_EC_ANDOR_NR;
-
-	value = g_key_file_get_boolean(config, "Headset", "VoiceRecognition",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_VOICE_RECOGNITION;
-
-	value = g_key_file_get_boolean(config, "Headset", "InBandRingtone",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_INBAND_RINGTONE;
-
-	value = g_key_file_get_boolean(config, "Headset", "VoiceTags",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_ATTACH_NUMBER_TO_VOICETAG;
-
-	value = g_key_file_get_boolean(config, "Headset", "RejectingCalls",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_REJECT_A_CALL;
-
-	value = g_key_file_get_boolean(config, "Headset", "EnhancedCallStatus",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_ENHANCES_CALL_STATUS;
-
-	value = g_key_file_get_boolean(config, "Headset", "EnhancedCallControl",
-					&err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_ENHANCES_CALL_CONTROL;
-
-	value = g_key_file_get_boolean(config, "Headset",
-					"ExtendedErrorResultCodes", &err);
-	if (err) {
-		debug("audio.conf: %s", err->message);
-		g_error_free(err);
-		err = NULL;
-	} else if (value)
-		ag_features |= AG_FEATURE_EXTENDED_ERROR_RESULT_CODES;
-
-	return ag_features;
+	return ag.features;
 }
 
 static gboolean hs_dc_timeout(struct audio_device *dev)
@@ -1780,16 +2083,23 @@ int headset_connect_rfcomm(struct audio_device *dev, GIOChannel *io)
 	return hs->tmp_rfcomm ? 0 : -EINVAL;
 }
 
-int headset_close_rfcomm(struct audio_device *dev)
+int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
+{
+	struct headset *hs = dev->headset;
+
+	if (hs->sco)
+		return -EISCONN;
+
+	hs->sco = io;
+
+	return 0;
+}
+
+static int headset_close_rfcomm(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
 	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
 
-	if (hs->ring_timer) {
-		g_source_remove(hs->ring_timer);
-		hs->ring_timer = 0;
-	}
-
 	if (rfcomm) {
 		g_io_channel_close(rfcomm);
 		g_io_channel_unref(rfcomm);
@@ -1807,6 +2117,14 @@ void headset_set_authorized(struct audio_device *dev)
 {
 	struct headset *hs = dev->headset;
 
+	/* For HFP telephony isn't ready just disconnect */
+	if (hs->hfp_active && !ag.telephony_ready) {
+		error("Unable to accept HFP connection since the telephony "
+				"subsystem isn't initialized");
+		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
+		return;
+	}
+
 	hs->rfcomm = hs->tmp_rfcomm;
 	hs->tmp_rfcomm = NULL;
 
@@ -1816,6 +2134,7 @@ void headset_set_authorized(struct audio_device *dev)
 
 	hs->auto_dc = FALSE;
 
+	/* For HSP (no special SLC setup) move to CONNECTED state */
 	if (!hs->hfp_active)
 		headset_set_state(dev, HEADSET_STATE_CONNECTED);
 }
@@ -1823,38 +2142,59 @@ void headset_set_authorized(struct audio_device *dev)
 void headset_set_state(struct audio_device *dev, headset_state_t state)
 {
 	struct headset *hs = dev->headset;
+	gboolean value;
 
 	if (hs->state == state)
 		return;
 
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
+		value = FALSE;
 		close_sco(dev);
 		headset_close_rfcomm(dev);
 		g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Disconnected",
 						DBUS_TYPE_INVALID);
+		dbus_connection_emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Connected",
+						DBUS_TYPE_BOOLEAN, &value);
+		telephony_device_disconnected(dev);
+		active_devices = g_slist_remove(active_devices, dev);
 		break;
 	case HEADSET_STATE_CONNECT_IN_PROGRESS:
 		break;
 	case HEADSET_STATE_CONNECTED:
 		close_sco(dev);
 		if (hs->state < state) {
+			value = TRUE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Connected",
 						DBUS_TYPE_INVALID);
+			dbus_connection_emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Connected",
+						DBUS_TYPE_BOOLEAN, &value);
+			active_devices = g_slist_append(active_devices, dev);
+			telephony_device_connected(dev);
 		} else if (hs->state == HEADSET_STATE_PLAYING) {
+			value = FALSE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Stopped",
 						DBUS_TYPE_INVALID);
+			dbus_connection_emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Playing",
+						DBUS_TYPE_BOOLEAN, &value);
 		}
 		break;
 	case HEADSET_STATE_PLAY_IN_PROGRESS:
 		break;
 	case HEADSET_STATE_PLAYING:
+		value = TRUE;
 		hs->sco_id = g_io_add_watch(hs->sco,
 					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 					(GIOFunc) sco_cb, dev);
@@ -1862,6 +2202,10 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_HEADSET_INTERFACE,
 						"Playing", DBUS_TYPE_INVALID);
+		dbus_connection_emit_property_changed(dev->conn, dev->path,
+						AUDIO_HEADSET_INTERFACE,
+						"Playing",
+						DBUS_TYPE_BOOLEAN, &value);
 
 		if (hs->sp_gain >= 0)
 			headset_send(hs, "\r\n+VGS=%u\r\n", hs->sp_gain);
@@ -1957,3 +2301,140 @@ int headset_get_sco_fd(struct audio_device *dev)
 
 	return g_io_channel_unix_get_fd(hs->sco);
 }
+
+int telephony_event_ind(int index)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	if (!ag.er_ind) {
+		debug("telephony_report_event called but events are disabled");
+		return -EINVAL;
+	}
+
+	send_foreach_headset(active_devices, hfp_cmp,
+				"\r\n+CIEV:%d,%d\r\n", index + 1,
+				ag.indicators[index].val);
+
+	return 0;
+}
+
+int telephony_response_and_hold_ind(int rh)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	ag.rh = rh;
+
+	/* If we aren't in any response and hold state don't send anything */
+	if (ag.rh < 0)
+		return 0;
+
+	send_foreach_headset(active_devices, hfp_cmp, "\r\n+BTRH:%d\r\n",
+				ag.rh);
+
+	return 0;
+}
+
+int telephony_incoming_call_ind(const char *number, int type)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	if (ag.ring_timer) {
+		debug("telephony_incoming_call_ind: already calling");
+		return -EBUSY;
+	}
+
+	g_free(ag.number);
+	ag.number = g_strdup(number);
+	ag.number_type = type;
+
+	ring_timer_cb(NULL);
+	ag.ring_timer = g_timeout_add(RING_INTERVAL, ring_timer_cb, NULL);
+
+	return 0;
+}
+
+int telephony_calling_stopped_ind(void)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	if (!ag.ring_timer)
+		return -EINVAL;
+
+	g_source_remove(ag.ring_timer);
+	ag.ring_timer = 0;
+
+	return 0;
+}
+
+int telephony_ready_ind(uint32_t features,
+			const struct indicator *indicators, int rh)
+{
+	ag.telephony_ready = TRUE;
+	ag.features = features;
+	ag.indicators = indicators;
+	ag.rh = rh;
+
+	debug("Telephony plugin initialized");
+
+	print_ag_features(ag.features);
+
+	return 0;
+}
+
+int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
+					int mprty, const char *number,
+					int type)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	if (number)
+		send_foreach_headset(active_devices, hfp_cmp,
+					"\r\n+CLCC:%d,%d,%d,%d,%d,%s,%d\r\n",
+					idx, dir, status, mode, mprty,
+					number, type);
+	else
+		send_foreach_headset(active_devices, hfp_cmp,
+					"\r\n+CLCC:%d,%d,%d,%d,%d\r\n",
+					idx, dir, status, mode, mprty);
+
+	return 0;
+}
+
+int telephony_subscriber_number_ind(const char *number, int type, int service)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	send_foreach_headset(active_devices, hfp_cmp,
+				"\r\n+CNUM:,%s,%d,,%d\r\n",
+				number, type, service);
+
+	return 0;
+}
+
+static int cwa_cmp(struct headset *hs)
+{
+	if (!hs->hfp_active)
+		return -1;
+
+	if (hs->cwa_enabled)
+		return 0;
+	else
+		return -1;
+}
+
+int telephony_call_waiting_ind(const char *number, int type)
+{
+	if (!active_devices)
+		return -ENODEV;
+
+	send_foreach_headset(active_devices, cwa_cmp, "\r\n+CCWA:%s,%d\r\n",
+				number, type);
+
+	return 0;
+}
